package raytracer;

import java.util.function.Consumer;

import org.lwjgl.input.Keyboard;
import org.lwjgl.input.Mouse;
import org.lwjgl.util.vector.Matrix4f;
import org.lwjgl.util.vector.Vector3f;

import raytracer.graphics.Shader;
import raytracer.math.MathUtil;

public class Camera {
	
	public float fov, aspect, near, far;
	public Vector3f position;
	public Vector3f rotation;
	public Matrix4f projectionMatrix, viewMatrix, invProjMatr, invViewMatr;
	
	public boolean moveWithMouse = false, M_down = false;
	
	private Consumer<Shader> updateUniforms = s -> {
		s.setUniformMat4("invViewMatr", invViewMatr);
		s.setUniform3f("position", position.x, position.y, position.z);
	};
	private Consumer<Shader> setInvProjMatrix = s -> s.setUniformMat4("invProjMatr", invProjMatr);
	
	public Camera(float fov, float aspect, float near, float far) {
		this.fov = fov;
		this.aspect = aspect;
		this.near = near;
		this.far = far;
		position = new Vector3f();
		rotation = new Vector3f();
		projectionMatrix = MathUtil.projectionMatrix(fov, aspect, this.near, this.far, null);
		invProjMatr = Matrix4f.invert(projectionMatrix, null);
		viewMatrix = MathUtil.viewMatrix(this, null);
		invViewMatr = Matrix4f.invert(viewMatrix, null);
	}
	
	public void update() {
		float speed = Keyboard.isKeyDown(Keyboard.KEY_LCONTROL) ? 0.01f : Keyboard.isKeyDown(Keyboard.KEY_LMENU) ? 0.25f : 0.05f, rSpeed = 0.025f;
		
		if (Keyboard.isKeyDown(Keyboard.KEY_W)) {
			position.z -= (float) (Math.cos(rotation.y) * speed);
			position.x += (float) (Math.sin(rotation.y) * speed);
		}
		if (Keyboard.isKeyDown(Keyboard.KEY_S)) {
			position.z += (float) (Math.cos(rotation.y) * speed);
			position.x -= (float) (Math.sin(rotation.y) * speed);
		}
		if (Keyboard.isKeyDown(Keyboard.KEY_D)) {
			position.x += (float) (Math.cos(rotation.y) * speed);
			position.z += (float) (Math.sin(rotation.y) * speed);
		}
		if (Keyboard.isKeyDown(Keyboard.KEY_A)) {
			position.x -= (float) (Math.cos(rotation.y) * speed);
			position.z -= (float) (Math.sin(rotation.y) * speed);
		}
		if (Keyboard.isKeyDown(Keyboard.KEY_SPACE)) {
			position.y += speed;
		}
		if (Keyboard.isKeyDown(Keyboard.KEY_LSHIFT)) {
			position.y -= speed;
		}
		
		if (Keyboard.isKeyDown(Keyboard.KEY_UP)) {
			rotation.x -= rSpeed;
			if (rotation.x < -Math.PI / 2) rotation.x = -(float) (Math.PI / 2);
		}
		if (Keyboard.isKeyDown(Keyboard.KEY_DOWN)) {
			rotation.x += rSpeed;
			if (rotation.x > Math.PI / 2) rotation.x = (float) (Math.PI / 2);
		}
		if (Keyboard.isKeyDown(Keyboard.KEY_LEFT)) {
			rotation.y -= rSpeed;
			if (rotation.y < 0) rotation.y += (float) (Math.PI * 2);
		}
		if (Keyboard.isKeyDown(Keyboard.KEY_RIGHT)) {
			rotation.y += rSpeed;
			if (rotation.y > Math.PI * 2) rotation.y -= (float) (Math.PI * 2);
		}
		if (Keyboard.isKeyDown(Keyboard.KEY_PRIOR)) {
			rotation.z -= rSpeed;
			if (rotation.z < 0) rotation.z += (float) (Math.PI * 2);
		}
		if (Keyboard.isKeyDown(Keyboard.KEY_NEXT)) {
			rotation.z += rSpeed;
			if (rotation.z > Math.PI * 2) rotation.z -= (float) (Math.PI * 2);
		}
		
		if (Keyboard.isKeyDown(Keyboard.KEY_M)) {
			if (!M_down) {
				M_down = true;
				moveWithMouse = !moveWithMouse;
				Mouse.setGrabbed(moveWithMouse);
			}
		} else M_down = false;
		
		if (moveWithMouse) {
			int dx = Mouse.getDX(), dy = Mouse.getDY();
			rotation.y += dx * rSpeed * 0.1f;
			rotation.x -= dy * rSpeed * 0.1f;
			
			if (rotation.x < -Math.PI / 2) rotation.x = -(float) (Math.PI / 2);
			else if (rotation.x > Math.PI / 2) rotation.x = (float) (Math.PI / 2);
			if (rotation.y < 0) rotation.y += (float) (Math.PI * 2);
			else if (rotation.y > Math.PI * 2) rotation.y -= (float) (Math.PI * 2);
		}
		
		MathUtil.viewMatrix(this, viewMatrix);
		Matrix4f.invert(viewMatrix, invViewMatr);
		Shader.forEachShader(updateUniforms);
	}
	
	public void setPosition(float x, float y, float z) {
		position.x = x;
		position.y = y;
		position.z = z;
	}
	
	public void loadInvProjMatrToShader(Shader shader) {
		if (shader == null) return;
		shader.enable();
		setInvProjMatrix.accept(shader);
	}
	
}
