package raytracer.math;

import org.lwjgl.util.vector.Matrix4f;
import org.lwjgl.util.vector.Vector3f;

import raytracer.Camera;

public class MathUtil {
	
	public static final Vector3f unitX = new Vector3f(1, 0, 0);
	public static final Vector3f unitY = new Vector3f(0, 1, 0);
	public static final Vector3f unitZ = new Vector3f(0, 0, 1);
	
	private MathUtil() {}
	
	public static Matrix4f projectionMatrix(float fov, float aspect, float near, float far, Matrix4f matr) {
		if (matr == null) matr = new Matrix4f();
		else matr.setIdentity();
		float y_scale = (float) (1 / Math.tan(Math.PI / 180 * fov / 2) * aspect);
		float frustLen = near - far;
		
		matr.m00 = y_scale / aspect;
		matr.m11 = y_scale;
		matr.m22 = (near + far) / frustLen;
		matr.m23 = -1;
		matr.m32 = 2 * near * far / frustLen;
		matr.m33 = 0;
		
		return matr;
	}
	
	public static Matrix4f viewMatrix(Camera camera, Matrix4f matr) {
		if (matr == null) matr = new Matrix4f();
		else matr.setIdentity();
		Matrix4f.rotate(camera.rotation.z, unitZ, matr, matr);
		Matrix4f.rotate(camera.rotation.x, unitX, matr, matr);
		Matrix4f.rotate(camera.rotation.y, unitY, matr, matr);
		return matr;
	}
	
	public static float hypot(float a, float b, float c) {
		return (float) Math.sqrt(a * a + b * b + c * c);
	}
	
	public static float clamp(float a, float low, float high) {
		return a < low ? low : a > high ? high : a;
	}
	
	public static int superMod(int x, int m) {
		x %= m;
		if (x < 0) x += m;
		return x;
	}
	
	public static float superMod(float x, float m) {
		x %= m;
		if (x < 0) x += m;
		if (x == m) x = 0;
		return x;
	}
	
	public static double superMod(double x, double m) {
		x %= m;
		if (x < 0) x += m;
		if (x == m) x = 0;
		return x;
	}
	
	public static void floorAll(Vector3f vec) {
		vec.x = (float) Math.floor(vec.x);
		vec.y = (float) Math.floor(vec.y);
		vec.z = (float) Math.floor(vec.z);
	}
	
}
