#version 330 core
#define PI 3.14159265358979

out vec4 color;

uniform vec2 windowSize;
uniform mat4 invProjMatr;
uniform mat4 invViewMatr;
uniform vec3 position;
uniform float time;// In seconds

vec3 gradF(vec3, float);

/**
    This is the graphics-file for RayTracer.

    If you run the program it will reload this file when it is saved. It will take less than 5 seconds.
    This means that you can edit the function and see the results within 5 seconds!

    Made by Martomate.
*/


/** This is the function f.
    When f(pos) = 0 there will be color on the screen.
    This is a contour-surface.

    Tip: If you want a normal function (y = g(x, z)) then write g(x, z) - y
*/
float f(vec3 pos) {
//    return dot(pos, pos) - 4;// sphere with radius 2
//    return sin(sqrt(pos.x * pos.x + pos.z * pos.z)) * (sin(pos.x + pos.z)) - pos.y - 10;
//    return pos.x * pos.x - pos.z * pos.z - pos.y - 10;
    float halfdx = 4;
    float x1 = pos.x - halfdx, x2 = pos.x + halfdx, zz = pos.z - 10;
    float velocity = 2;
    float speed = 1;
    return (sin(velocity * 1.4 * sqrt(x1 * x1 + zz * zz) - 2 * PI * time * speed) + sin(velocity * sqrt(x2 * x2 + zz * zz) - 2 * PI * time * speed)) / 2 - pos.y - 2;
}

/** The colorcoding used at a position pos */
vec3 colorCode(vec3 pos) {
//    vec3 grad = normalize(gradF(pos, 0.01));return abs(grad);
    float amp = pos.y + 2;return vec3(0, (amp + 1) / 2, (amp + 3) / 4);
}


/** The rest of the code */// ------------------------------------------------------------------------------

vec3 gradF(vec3 pos, float dt) {
    vec3 dx = vec3(dt, 0, 0);
    vec3 dy = vec3(0, dt, 0);
    vec3 dz = vec3(0, 0, dt);
    float diffx = f(pos + dx) - f(pos - dx);
    float diffy = f(pos + dy) - f(pos - dy);
    float diffz = f(pos + dz) - f(pos - dz);
    return vec3(diffx, diffy, diffz) / (2 * dt);
}

void main() {
	vec4 coordsBeforeMatr = vec4(gl_FragCoord.x / windowSize.x * 2 - 1, gl_FragCoord.y / windowSize.y * 2 - 1, -1, 1);
	vec4 coordsAfterProj = invProjMatr * coordsBeforeMatr;
	vec3 ray = normalize((invViewMatr * vec4(coordsAfterProj.xy, -1, 0)).xyz);
	vec3 sun = normalize(ray);

	vec3 pos = position;
	float fBefore = f(position);

	const int iterations = 100;
	const float rho = 1.0;
	float dt = rho * 0.1;
	float limit = 0.001;
	float value = fBefore;
	const int iter10 = iterations * 10;
	const float scale = 1.005;
    for (int i = 0; i < iter10; i++) {
        value = f(pos);
        pos += ray * dt;
        if (value * fBefore < 0) {
            pos -= ray * dt;
            i = iter10;
        }
        dt *= scale;
    }
    dt = rho * 0.1;
	for (int i = 0; i < iterations; i++) {
	    vec3 smallStep = ray * (dt * 0.01);
	    float dfdt = (f(pos + smallStep) - f(pos - smallStep)) / (dt * 0.02);
	    value = f(pos);
		if (dfdt != 0) {
		    float newDt = -rho * value / dfdt;
		    if (newDt < 0 && value * fBefore > 0) newDt *= -1;
		    else if (newDt * dt < 0) newDt *= 0.5;
		    dt = newDt;
		}
		pos += ray * dt;
		if (abs(value) < limit) i = iterations;
	}

	value = f(pos);
    float intensity = 1.0 / (1 + value * value);

	color = vec4(colorCode(pos) * intensity, 1);
}
