#version 330 core

out vec4 color;

uniform vec2 windowSize;
uniform mat4 invProjMatr;
uniform mat4 invViewMatr;
uniform vec3 position;

float f(vec3 pos) {
//    return dot(pos, pos) - 4;// sphere with radius 2
    return sin(sqrt(pos.x * pos.x + pos.z * pos.z)) * (1 + abs(pos.x + pos.z)) - pos.y;
}

vec3 gradF(vec3 pos, float dt) {
    vec3 dx = vec3(dt, 0, 0);
    vec3 dy = vec3(0, dt, 0);
    vec3 dz = vec3(0, 0, dt);
    float diffx = f(pos + dx) - f(pos - dx);
    float diffy = f(pos + dy) - f(pos - dy);
    float diffz = f(pos + dz) - f(pos - dz);
    return vec3(diffx, diffy, diffz) / (2 * dt);
}

void main() {
	vec4 coordsBeforeMatr = vec4(gl_FragCoord.x / windowSize.x * 2 - 1, gl_FragCoord.y / windowSize.y * 2 - 1, -1, 1);
	vec4 coordsAfterProj = invProjMatr * coordsBeforeMatr;
	vec3 ray = normalize((invViewMatr * vec4(coordsAfterProj.xy, -1, 0)).xyz);
	vec3 sun = normalize(ray);

	vec3 pos = position;
	float fBefore = f(position);

	const int iterations = 100;
	const float rho = 1;
	float dt = rho * 0.1;
	float dtLimit = 0.0001;
	float value = fBefore;
	for (int i = 0; i < iterations; i++) {
        value = f(pos);
        pos += ray * dt;
        if (value * fBefore < 0) {
            i = iterations;
            pos -= ray * dt;
            dt /= 10;
        }
    }
	for (int i = 0; i < iterations; i++) {
	    vec3 smallStep = ray * (dt * 0.1);
	    float dfdt = (f(pos + smallStep) - f(pos - smallStep)) / (dt * 0.2);
	    value = f(pos);
		if (dfdt != 0) {
		    dt = -rho * value / dfdt;
		    if (dt < 0 && value * fBefore > 0) dt *= 0;
		}
		pos += ray * dt;
		if (abs(dt) < dtLimit) i = iterations;
	}
	value = f(pos);
	float intensity = 1.0 / (1 + value * value);
	vec3 grad = normalize(gradF(pos, 0.01));
	color = vec4(abs(grad) * intensity, 1);
}
